'' =================================================================================================
''
''   File....... jm_ws2801.spin
''   Purpose.... High-performance driver for the WS2801
''   Author..... Jon "JonnyMac" McPhalen 
''               Copyright (c) 2012 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 25 JUL 2012
''
'' =================================================================================================

{

  Recommended frame timing (@ 80MHz) is: (10us x channels) + 1ms

}


con

  RX1 = 31                                                      ' programming / terminal
  TX1 = 30
  
  SDA = 29                                                      ' eeprom / i2c
  SCL = 28

  
con

  MAX_CHANNELS = 64                                             ' adjust as required
  

var

  long  cog                                                     ' id of cog running driver

  long  dpin                                                    ' -> SDI pin
  long  cpin                                                    ' -> CKI pin
  long  chcount                                                 ' channel count in frame
  long  frametix                                                ' frame timing (in cnt ticks)
  
  long  color[MAX_CHANNELS]                                     ' color array


pub start(d, c, ch, ms)

'' Start multi-channel WS2801 driver
'' -- d is pin that connects to sdi (serial data in)
'' -- c is pin that connects to cki (clock in)
'' -- ch is the number of channels (RGB groups) in string
'' -- ms is the frame refresh timing in milliseconds  

  stop

  dpin     := d                                                 ' set pins
  cpin     := c
  chcount  := 1 #> ch <# MAX_CHANNELS                           ' set channels (w/limits)
  frametix := clkfreq / 1_000 * ms                              ' ticks in frame
  
  cog := cognew(@ws2801, @dpin) + 1                             ' start the cog

  return cog


pub stop

'' Stops cog (if running)

  if (cog)
    cogstop(cog - 1)
    cog := 0

  longfill(@color, 0, MAX_CHANNELS)                             ' clear all to off
  

pub set(ch, rgb) : work

'' Sets channel to RGB value
'' -- re-orders bytes to accomodate RGB value (e.g., $FF_FF_00 = yellow)

  if ((ch => 0) and (ch < MAX_CHANNELS))                        ' valid?
    work.byte[2] := rgb.byte[0]                                 '  re-arrange for strip
    work.byte[1] := rgb.byte[1]
    work.byte[0] := rgb.byte[2]
    color[ch] := work                                           '  update color array


pub set_red(ch, level)

'' Sets red led level of selected channel

  if ((ch => 0) and (ch < MAX_CHANNELS))                        ' valid?
    byte[@color + (ch << 2) + 0] := 0 #> level <# 255           '  set it
    

pub set_green(ch, level)

'' Sets green led level of selected channel

  if ((ch => 0) and (ch < MAX_CHANNELS))                        ' valid?
    byte[@color + (ch << 2) + 1] := 0 #> level <# 255           '  set it


pub set_blue(ch, level)

'' Sets blue led level of selected channel

  if ((ch => 0) and (ch < MAX_CHANNELS))                        ' valid?
    byte[@color + (ch << 2) + 2] := 0 #> level <# 255           '  set it

    
pub set_all(rgb) : work

'' Sets all leds in string to RGB

  work.byte[2] := rgb.byte[0]                                   ' re-arrange for strip
  work.byte[1] := rgb.byte[1] 
  work.byte[0] := rgb.byte[2]
  
  longfill(@color, work, MAX_CHANNELS)


pub address

'' Returns address of color array

  return @color
  

pub running

'' Returns > 0 if running

  return cog
 

dat

                        org     0

ws2801                  mov     t1, par                         ' copy hub address of parameters
                        rdlong  sdipin, t1                      ' read sdi pin #
                        add     t1, #4
                        rdlong  ckipin, t1                      ' read cki pin #
                        add     t1, #4
                        rdlong  nleds, t1                       ' read # of channels
                        add     t1, #4
                        rdlong  frame, t1                       ' read frame timing
                        add     t1, #4
                        mov     hub, t1                         ' save hub address of colors

                        mov     sdimask, #1                     ' convert sdi pin to mask
                        shl     sdimask, sdipin
                        or      dira, sdimask                   ' make output

                        mov     ckimask, #1
                        shl     ckimask, ckipin
                        or      dira, ckimask

                        mov     frametimer, frame               ' start and sync frame timer
                        add     frametimer, cnt
                        
rgbmain                 mov     pntr, hub
                        mov     count, nleds

getrgb                  rdlong  t1, pntr

shiftout                shl     t1, #8                          ' align msb
                        mov     bits, #24                       ' 24 bits = 3 rgb bytes
:loop                   rcl     t1, #1                  wc      ' move msb to C
                        muxc    outa, sdimask                   ' copy C to sdi 
                        nop                                     ' let sdi stabilize
                        or      outa, ckimask                   ' clock the bit
                        nop
                        andn    outa, ckimask
                        djnz    bits, #:loop                    ' update bit count, do next
                        
                        add     pntr, #4                        ' point to next rgb value
                        djnz    count, #getrgb                  ' update count, get next

                        waitcnt frametimer, frame               ' let frame finish
                        jmp     #rgbmain

' --------------------------------------------------------------------------------------------------

sdipin                  res     1                               ' SDI (serial data) pin
ckipin                  res     1                               ' CKI (clock) pin
nleds                   res     1                               ' number of RGB leds in chain
frame                   res     1                               ' frame timing
hub                     res     1                               ' hub address of color[] array

sdimask                 res     1                               ' pin mask for SDI
ckimask                 res     1                               ' pin mask for CKI
pntr                    res     1                               ' data pointer
count                   res     1                               ' count of leds to process
bits                    res     1                               ' bits in shiftout

frametimer              res     1                               ' for frame timing

t1                      res     1                               ' work vars
t2                      res     1

                        fit     496                                    

                        
dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}  